# The MIT License (MIT)
#
# Copyright (c) 2016 Philippe Proulx <pproulx@efficios.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from lamiview.dialog_progress import QLamiViewProgressDialog
from lamiview.main_window import QLamiViewMainWindow
from collections import namedtuple
from PyQt5 import Qt, QtCore
import pygments.lexers.data
import pygments.formatters
from lamiview import lami
import pygments
import argparse
import platform
import pygments
import logging
import json
import sys
import os
import re


_logger = logging.getLogger(__name__)


_Config = namedtuple('_Config', ['commands'])


class _QLamiViewApp(Qt.QApplication):
    _UPDATE_TIMER_MS = 1000

    def __init__(self, args):
        super().__init__([sys.argv[0]])
        self._args = args
        self.setOrganizationName('eepp.ca')
        self.setApplicationName('lamiview')
        self._setup()
        self._start()

    def _start(self):
        _logger.info('Starting application')
        self._main_window.show()

    def _setup(self):
        self._load_config()
        self._setup_ui()
        self._setup_quit_timer()
        self._setup_actions()

    def _setup_ui(self):
        self._main_window = QLamiViewMainWindow()
        self._progress_dialog = QLamiViewProgressDialog()
        self._init_input()

    def _init_input(self):
        for command in self._config.commands:
            self._main_window.add_command(command)

        if self._args.cmd:
            self._main_window.set_command(self._args.cmd)

        if self._args.path:
            self._main_window.set_path(self._args.path[0])

        if self._args.begin:
            self._main_window.set_begin(self._args.begin)

        if self._args.end:
            self._main_window.set_end(self._args.end)

        if self._args.limit:
            self._main_window.set_limit(self._args.limit)

        if self._args.output_progress:
            self._main_window.set_output_progress(self._args.output_progress)

    def _show_progress_dialog(self):
        pos = self._main_window.pos()
        pos.setX(pos.x() + 40)
        pos.setY(pos.y() + 40)
        self._progress_dialog.reset()
        self._progress_dialog.set_value(.45)
        self._progress_dialog.set_message('29312 events processed')
        self._progress_dialog.show_move(pos)

    def _setup_actions(self):
        self._main_window.on_action_quit_triggered(self.closeAllWindows)
        self._main_window.on_btn_metadata_clicked(self._get_metadata)
        self._main_window.on_btn_analysis_clicked(self._get_analysis)

    def _setup_quit_timer(self):
        self._quit_timer = Qt.QTimer()
        self._quit_timer.timeout.connect(lambda: None)
        self._quit_timer.start(250)

    def _load_config(self):
        self._config = _Config(commands=[])
        config_path = os.getenv('LAMIVIEW_CONFIG_PATH',
                                os.path.expanduser('~/.config/lamiview/config.json'))

        if not os.path.isfile(config_path):
            _logger.warning('Config file "{}" is not an existing file'.format(config_path))
            return

        try:
            with open(config_path, 'r') as f:
                json_config = json.load(f)
        except Exception as e:
            _logger.warning('Cannot open/parse config file "{}": {}'.format(config_path, e))
            return

        _logger.info('Loaded config file "{}"'.format(config_path))
        commands = []

        if 'commands' in json_config and type(json_config['commands']) is list:
            for command in json_config['commands']:
                if type(command) is str:
                    _logger.info('Adding command "{}" from config file'.format(command))
                    commands.append(command)
                else:
                    _logger.warning('Not adding invalid command "{}" from config file'.format(command))

        self._config = _Config(commands=commands)

    def _show_status_message(self, msg):
        self._main_window.show_status_message(msg)

    def _clear_status_message(self):
        self._main_window.clear_status_message()

    def _dict_to_json_html(self, dictionary):
        json_str = json.dumps(dictionary, indent=2, sort_keys=True)
        formatter = pygments.formatters.HtmlFormatter(noclasses=True,
                                                      nobackground=True)
        lexer = pygments.lexers.data.JsonLexer()
        json_html = pygments.highlight(json_str, lexer, formatter)
        html = '''
            <!doctype html>
            <html>
              <head>
                <meta charset="utf-8">
                <title></title>
                <style type="text/css">
                  html {{
                    font-size: 8pt;
                  }}

                  body, div, pre {{
                    margin: 0;
                    padding: 0;
                  }}

                  pre {{
                    padding: 1em;
                  }}
                </style>
              </head>
              <body>{}</body>
            </html>'''.format(json_html)

        return html

    def _create_base_table_widget(self):
        table = Qt.QTableWidget()
        table.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)
        table.setStyleSheet('''
            QWidget {
              font-size: 7pt;
              font-family: monospace;
            }

            QHeaderView {
              font-weight: bold;
              font-family: sans-serif;
            }''')

        return table

    def _create_table_widget_item(self, text):
        item = Qt.QTableWidgetItem(text)
        item.setFlags(QtCore.Qt.NoItemFlags | QtCore.Qt.ItemIsEnabled)

        return item

    def _fix_table_widget(self, table):
        table.verticalHeader().setDefaultSectionSize(15)
        table.resizeColumnsToContents()
        hheader = table.horizontalHeader()

        for index in range(hheader.count()):
            section_size = hheader.sectionSize(index)
            section_size += 6
            hheader.resizeSection(index, section_size)

        height = (table.rowCount()) * 15 + table.horizontalHeader().height() + 1
        height += table.horizontalScrollBar().height()
        table.setMaximumHeight(height)
        table.setMinimumHeight(height)

    def _create_metadata_infos_table(self, metadata):
        rows = []
        rows.append(('MI version', str(metadata.mi_version)))

        if metadata.version is not None:
            rows.append(('Version', str(metadata.version)))

        if metadata.title is not None:
            rows.append(('Title', str(metadata.title)))

        if metadata.authors:
            rows.append(('Authors', ', '.join(metadata.authors)))

        if metadata.description is not None:
            rows.append(('Description', metadata.description))

        if metadata.url is not None:
            rows.append(('URL', metadata.url))

        if metadata.tags:
            rows.append(('Tags', ', '.join(metadata.tags)))

        table = self._create_base_table_widget()
        table.setRowCount(len(rows))
        table.setColumnCount(2)
        table.setHorizontalHeaderLabels(('Property', 'Value'))

        for row_index, row in enumerate(rows):
            prop = self._create_table_widget_item(row[0])
            value = self._create_table_widget_item(row[1])
            table.setItem(row_index, 0, prop)
            table.setItem(row_index, 1, value)

        self._fix_table_widget(table)

        return table

    def _wrap_table_horizontally(self, table):
        return table

    def _create_metadata_table_class_table(self, table_class):
        table = self._create_base_table_widget()
        table.setRowCount(len(table_class.column_descriptions))
        table.setColumnCount(3)
        table.setHorizontalHeaderLabels(('Title', 'Class', 'Unit'))

        for row_index, cd in enumerate(table_class.column_descriptions):
            title = ''
            cls = 'mixed'
            unit = ''

            if cd.title:
                title = cd.title

            if cd.cls is not None:
                if type(cd.cls) is str:
                    cls = cd.cls
                else:
                    cls = cd.cls.NAME

            if cd.unit:
                unit = cd.unit

            title = self._create_table_widget_item(title)
            cls = self._create_table_widget_item(cls)
            unit = self._create_table_widget_item(unit)
            table.setItem(row_index, 0, title)
            table.setItem(row_index, 1, cls)
            table.setItem(row_index, 2, unit)

        self._fix_table_widget(table)

        return table

    def _set_metadata_output(self, metadata):
        # clear things
        self._main_window.set_raw_metadata_html('')
        self._main_window.clear_metadata_widgets()

        raw_metadata_html = self._dict_to_json_html(metadata.dict)
        self._main_window.set_raw_metadata_html(raw_metadata_html)

        # set interpreted metadata output
        if type(metadata) is lami.Error:
            return

        metadata_table = self._create_metadata_infos_table(metadata)
        widget = self._wrap_table_horizontally(metadata_table)
        self._main_window.add_metadata_widget('Infos', widget)

        for cls_name, table_class in metadata.table_classes.items():
            table = self._create_metadata_table_class_table(table_class)
            title = 'Table class <span style="color: #d35400;">{}</span>'.format(cls_name)
            title_widget = Qt.QLabel(title)
            title_widget.setStyleSheet('QLabel { font-size: 8pt; font-weight: bold; }')
            widget = self._wrap_table_horizontally(table)
            self._main_window.add_metadata_widget(title_widget, widget)

    def _get_metadata(self):
        # clear things
        self._main_window.clear_all()

        # get metadata
        self._show_status_message('Getting metadata...')
        self._main_window.enable_input_box(False)
        Qt.QCoreApplication.processEvents()

        try:
            metadata = lami.get_metadata(self._main_window.command)
        except Exception as e:
            self._main_window.set_raw_metadata_html(self._get_html_error(str(e)))
            self._main_window.focus_raw_metadata()
            return
        finally:
            self._clear_status_message()
            self._main_window.enable_input_box(True)

        # set raw metadata output
        self._set_metadata_output(metadata)
        self._main_window.focus_metadata()

    def _create_analysis_result_table(self, result_table):
        table = self._create_base_table_widget()
        table.setRowCount(len(result_table.rows))
        table.setColumnCount(len(result_table.cls.column_descriptions))
        labels = []

        for index, cd in enumerate(result_table.cls.column_descriptions):
            def make_label(parts):
                if unit:
                    parts.append('({})'.format(unit))

                return ' '.join(parts)

            if cd.title:
                title = cd.title
            else:
                title = 'Column #{}'.format(index + 1)

            unit = None

            if cd.unit:
                unit = cd.unit
            else:
                if hasattr(cd.cls, 'UNIT') and cd.cls.UNIT:
                    unit = cd.cls.UNIT

            if cd.cls.COLUMNS:
                for col_name, col_attr in cd.cls.COLUMNS:
                    labels.append(make_label([title, '[{}]'.format(col_name)]))
            else:
                labels.append(make_label([title]))

        table.setHorizontalHeaderLabels(tuple(labels))

        for row_index, row in enumerate(result_table.rows):
            cells = []

            for col_index, cell in enumerate(row):
                cd = result_table.cls.column_descriptions[col_index]

                if cd.cls.COLUMNS:
                    for col_name, col_attr in cd.cls.COLUMNS:
                        if type(cell) is lami.Unknown:
                            cells.append('?')
                        elif type(cell) is lami.Empty:
                            cells.append('')
                        else:
                            attr_value = getattr(cell, col_attr)
                            cells.append(str(attr_value))
                else:
                    cells.append(str(cell))

            for col_index, cell in enumerate(cells):
                item = self._create_table_widget_item(cell)
                table.setItem(row_index, col_index, item)

        self._fix_table_widget(table)

        return table

    def _get_analysis(self):
        def on_progress_update(progress):
            if progress.message:
                self._progress_dialog.set_message(progress.message)

            self._progress_dialog.set_value(progress.value)
            Qt.QCoreApplication.processEvents()

        # clear things
        self._main_window.clear_all()

        # get analysis results
        command = self._main_window.command
        path = self._main_window.path
        begin = self._main_window.begin
        end = self._main_window.end
        limit = self._main_window.limit
        output_progress = self._main_window.output_progress

        if output_progress:
            self._show_progress_dialog()

        self._show_status_message('Getting analysis results...')
        self._main_window.enable_input_box(False)
        Qt.QCoreApplication.processEvents()

        try:
            results = lami.get_analysis_results(command, path, begin, end, limit,
                                                output_progress, on_progress_update)
        except Exception as e:
            self._main_window.set_raw_analysis_html(self._get_html_error(str(e)))
            self._main_window.focus_raw_analysis()
            return
        finally:
            self._clear_status_message()
            self._progress_dialog.done(0)
            self._main_window.enable_input_box(True)

        # set raw analysis results output
        raw_analysis_html = self._dict_to_json_html(results.dict)
        self._main_window.set_raw_analysis_html(raw_analysis_html)

        # set interpreted analysis results output
        if type(results) is lami.Error:
            return

        # set metadata output
        self._set_metadata_output(results.metadata)

        for index, result_table in enumerate(results.result_tables):
            if result_table.cls.title:
                title = result_table.cls.title
            else:
                title = 'Result table #{}'.format(index + 1)

            title_widget = Qt.QLabel(title)
            title_widget.setTextFormat(QtCore.Qt.PlainText)
            title_widget.setStyleSheet('''
                QLabel {
                  color: #d35400;
                  font-size: 8pt;
                  font-weight: bold;
                }''')

            horiz_widget = Qt.QWidget()
            horiz_layout = Qt.QHBoxLayout(horiz_widget)
            horiz_layout.setContentsMargins(0, 0, 0, 0)
            horiz_layout.setSpacing(12)
            horiz_layout.addWidget(title_widget)
            horiz_layout.addStretch()

            if result_table.time_range:
                label = Qt.QLabel(str(result_table.time_range))
                label.setTextFormat(QtCore.Qt.PlainText)
                label.setStyleSheet('''
                    QLabel {
                      font-size: 8pt;
                      font-family: monospace;
                      font-weight: bold;
                    }''')
                horiz_layout.addWidget(label)

            widget = self._create_analysis_result_table(result_table)
            self._main_window.add_analysis_widget(horiz_widget, widget)

        self._main_window.focus_analysis()

    def _get_html_error(self, msg):
        return '''
            <!doctype html>
            <html>
              <head>
                <meta charset="utf-8">
                <title></title>
                <style type="text/css">
                  html {{
                    font-size: 8pt;
                  }}

                  body, div, p {{
                    margin: 0;
                    padding: 0;
                  }}

                  body {{
                    margin: 1em;
                    font-family: monospace;
                  }}

                  p {{
                    color: #c0392b;
                  }}
                </style>
              </head>
              <body>
                <p><strong>Error</strong> while executing the command:<br>{}</p>
              </body>
            </html>'''.format(msg)


def _register_sigint(app):
    if platform.system() == 'Linux':
        def handler(signal, frame):
            _logger.info('Got SIGINT')
            sys.exit(0)

        import signal

        signal.signal(signal.SIGINT, handler)


def _configure_logging():
    fmt = '%(asctime)s [%(levelname)7s] ' \
          '%(message)s (%(funcName)s@%(filename)s:%(lineno)d)'
    logging.basicConfig(level=logging.INFO, format=fmt)


def _parse_args():
    import lamiview

    ap = argparse.ArgumentParser()
    ap.add_argument('-V', '--version', action='version',
                    version='%(prog)s v{}'.format(lamiview.__version__))
    ap.add_argument('-b', '--begin', metavar='TS', action='store',
                    help='beginning timestamp')
    ap.add_argument('-e', '--end', metavar='TS', action='store',
                    help='ending timestamp')
    ap.add_argument('-l', '--limit', metavar='COUNT', action='store',
                    help='maximum rows/table')
    ap.add_argument('-p', '--output-progress', action='store_true',
                    help='show progress indication')
    ap.add_argument('-c', '--cmd', metavar='COMMAND', action='store',
                    help='command')
    ap.add_argument('path', metavar='PATH', action='store', nargs='*',
                    help='trace path')

    # parse args
    args = ap.parse_args()

    if len(args.path) > 1:
        _logger.error('Command-line error: Please specify zero or one trace path')
        return None

    return args


def run():
    args = _parse_args()

    if args is None:
        return 1

    _configure_logging()
    app = _QLamiViewApp(args)
    _register_sigint(app)
    rc = app.exec_()
    _logger.info('App returned exit status {}'.format(rc))

    return rc


if __name__ == '__main__':
    run()



from lamiview import __version__, utils
from PyQt5 import Qt
import os.path


class QLamiViewAboutDialog(utils.QCommonDialog, utils.QtUiLoad):
    _UI_NAME = 'dialog_about'

    def __init__(self):
        super().__init__()
        self._setup_ui()

    def _set_version(self):
        self._lbl_version.setText('v{}'.format(__version__))

    def _set_contents(self):
        self._set_version()

    def _setup_ui(self):
        self._load_ui()
        self._set_contents()



# The MIT License (MIT)
#
# Copyright (c) 2016 Philippe Proulx <pproulx@efficios.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import os.path
import logging
from PyQt5 import Qt, QtCore
from lamiview import utils
from lamiview.dialog_about import QLamiViewAboutDialog


_logger = logging.getLogger(__name__)


class QLamiViewMainWindow(Qt.QMainWindow, utils.QtUiLoad):
    _UI_NAME = 'main_window'

    def __init__(self):
        super().__init__()
        self._setup_ui()
        self._setup_dialogs()
        self._setup_actions()
        self._setup_signals()

    def _setup_dialogs(self):
        self._about_dialog = QLamiViewAboutDialog()

    def _setup_ui(self):
        self._load_ui()
        self._setup_scroll_widgets()
        self._combo_command.setFocus()

    def _create_scroll_layout(self, widget):
        cur_layout = widget.layout()

        if cur_layout:
            cur_layout.setParent(None)

        layout = Qt.QVBoxLayout(widget)
        layout.setSpacing(3)
        layout.setContentsMargins(6, 6, 6, 6)
        layout.setAlignment(QtCore.Qt.AlignTop)

        return layout

    def _setup_scroll_widgets(self):
        self._scroll_metadata_widget = Qt.QWidget()
        self._scroll_metadata_layout = self._create_scroll_layout(self._scroll_metadata_widget)
        self._scroll_metadata.setWidget(self._scroll_metadata_widget)
        self._scroll_analysis_widget = Qt.QWidget()
        self._scroll_analysis_layout = self._create_scroll_layout(self._scroll_analysis_widget)
        self._scroll_analysis.setWidget(self._scroll_analysis_widget)

    def _show_about_dialog(self):
        pos = self.pos()
        pos.setX(pos.x() + 40)
        pos.setY(pos.y() + 40)
        self._about_dialog.show_move(pos)

    def _goto_to_lami_spec(self):
        utils.goto_url('https://github.com/lttng/lttng-analyses/tree/master/doc')

    def _setup_actions(self):
        self._action_about.triggered.connect(self._show_about_dialog)
        self._action_lami_spec.triggered.connect(self._goto_to_lami_spec)

    def _setup_signals(self):
        pass

    def on_action_quit_triggered(self, fn):
        self._action_quit.triggered.connect(fn)

    def on_btn_metadata_clicked(self, fn):
        self._btn_metadata.clicked.connect(fn)

    def on_btn_analysis_clicked(self, fn):
        self._btn_analysis.clicked.connect(fn)

    def add_command(self, cmd):
        self._combo_command.addItem(cmd)

    def set_command(self, cmd):
        self._combo_command.setCurrentText(cmd)

    def set_path(self, path):
        self._edit_path.setText(path)

    def set_begin(self, begin):
        self._edit_begin.setText(begin)

    def set_end(self, end):
        self._edit_end.setText(end)

    def set_limit(self, limit):
        self._edit_limit.setText(limit)

    def set_output_progress(self, output_progress):
        self._check_output_progress.setChecked(output_progress)

    def set_raw_metadata_html(self, html):
        self._web_raw_metadata.setHtml(html)

    def set_raw_analysis_html(self, html):
        self._web_raw_analysis.setHtml(html)

    def _add_widget_to_scroll_layout(self, layout, title, widget, highlight):
        # remove stretch first
        stretch_item = layout.takeAt(layout.count() - 1)

        # title widget
        if isinstance(title, Qt.QWidget):
            title_widget = title
        else:
            title_widget = Qt.QLabel(title)
            title_widget.setTextFormat(QtCore.Qt.PlainText)
            add_css = ''

            if highlight:
                add_css = 'color: #d35400;'

            title_widget.setStyleSheet('QLabel {{ font-size: 8pt; font-weight: bold; {} }}'.format(add_css))

        layout.addWidget(title_widget)

        # widget
        layout.addWidget(widget)

        # spacer
        layout.addSpacing(12)

        # add stretch again
        layout.addItem(stretch_item)

    def add_metadata_widget(self, title, widget, highlight=False):
        self._add_widget_to_scroll_layout(self._scroll_metadata_layout,
                                          title, widget, highlight)


    def add_analysis_widget(self, title, widget, highlight=False):
        self._add_widget_to_scroll_layout(self._scroll_analysis_layout,
                                          title, widget, highlight)

    def _clear_scroll_layout(self, layout):
        while layout.count() > 0:
            item = layout.takeAt(0)

            if type(item) is Qt.QWidgetItem:
                item.widget().setParent(None)

        layout.addStretch()

    def clear_metadata_widgets(self):
        self._clear_scroll_layout(self._scroll_metadata_layout)

    def clear_analysis_widgets(self):
        self._clear_scroll_layout(self._scroll_analysis_layout)

    def clear_all(self):
        self.clear_metadata_widgets()
        self.clear_analysis_widgets()
        self.set_raw_metadata_html('')
        self.set_raw_analysis_html('')

    def show_status_message(self, msg):
        self._status_bar.showMessage(msg)

    def clear_status_message(self):
        self._status_bar.clearMessage()

    def enable_input_box(self, enable):
        self._gb_input.setEnabled(enable)

    def enable_output_box(self, enable):
        self._gb_output.setEnabled(enable)

    def focus_raw_metadata(self):
        self._tabs_output.setCurrentIndex(0)

    def focus_metadata(self):
        self._tabs_output.setCurrentIndex(1)

    def focus_raw_analysis(self):
        self._tabs_output.setCurrentIndex(2)

    def focus_analysis(self):
        self._tabs_output.setCurrentIndex(3)

    @property
    def command(self):
        return self._combo_command.currentText()

    @property
    def path(self):
        return self._edit_path.text()

    @property
    def begin(self):
        text = self._edit_begin.text()

        if not text:
            return

        return text

    @property
    def end(self):
        text = self._edit_end.text()

        if not text:
            return

        return text

    @property
    def limit(self):
        text = self._edit_limit.text()

        if not text:
            return

        return text

    @property
    def output_progress(self):
        return self._check_output_progress.isChecked()



from lamiview import __version__, utils
from PyQt5 import Qt
import os.path


class QLamiViewProgressDialog(Qt.QDialog, utils.QtUiLoad):
    _UI_NAME = 'dialog_progress'

    def __init__(self):
        super().__init__()
        self._setup_ui()

    def _setup_ui(self):
        self._load_ui()
        self.setModal(True)

        # hide this until we aborting is supported
        self._btn_abort.hide()

    def set_value(self, val):
        self._pb.setValue(int(val * 100))

    def set_message(self, msg):
        self._lbl_msg.setText(msg)

    def reset(self):
        self.set_value(0)
        self.set_message('')

    def closeEvent(self, event):
        event.ignore()

    def show_move(self, pos):
        self.move(pos)
        self.show()



# The MIT License (MIT)
#
# Copyright (c) 2016 Philippe Proulx <pproulx@efficios.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import json
import logging
import subprocess
from lamiview import utils
from functools import total_ordering


_logger = logging.getLogger(__name__)


class UnsupportedVersion(Exception):
    pass


class ExecError(Exception):
    def __init__(self, exit_status=None, output=None):
        self._exit_status = exit_status
        self._output = output

    @property
    def exit_status(self):
        return self._exit_status

    @property
    def output(self):
        return self._output

    def __str__(self):
        string = 'Execution error'

        if self._exit_status is not None:
            string += ': command returned exit status {}'.format(self._exit_status)

        return string


class ParseError(Exception):
    pass


class Error:
    def __init__(self, code, message):
        self._code = code
        self._message = message

    @property
    def code(self):
        return self._code

    @property
    def message(self):
        return self._message

    @classmethod
    def from_obj(cls, obj):
        msg = None
        code = None

        if 'error-code' in obj:
            code = obj['error-code']

        if 'error-message' in obj:
            msg = obj['error-message']

        return cls(code, msg)

    @property
    def dict(self):
        return self._dict


class _Data:
    UNIT = None
    NAME = None
    COLUMNS = None


class _SimpleValue(_Data):
    def __init__(self, value):
        super().__init__()
        self._value = value

    @property
    def value(self):
        return self._value

    @classmethod
    def from_obj(cls, obj):
        return cls(obj['value'])

    def __str__(self):
        if type(self) is float:
            return str(round(self._value, 3))

        return str(self._value)


class _SimpleName(_Data):
    def __init__(self, name):
        super().__init__()
        self._name = name

    @property
    def name(self):
        return self._name

    @classmethod
    def from_obj(cls, obj):
        return cls(obj['name'])

    def __str__(self):
        return str(self._name)


class Empty(_Data):
    CLASS = 'empty'
    NAME = 'Empty'

    def __str__(self):
        return ''


class Unknown(_Data):
    CLASS = 'unknown'
    NAME = 'Unknown'

    def __init__(self):
        super().__init__()

    @classmethod
    def from_obj(cls, obj):
        return cls()

    def __str__(self):
        return '?'


class Integer(_SimpleValue):
    CLASS = 'int'
    NAME = 'Integer'


class Number(_SimpleValue):
    CLASS = 'number'
    NAME = 'Number'


class String(_SimpleValue):
    CLASS = 'string'
    NAME = 'String'


class Bool(_SimpleValue):
    CLASS = 'bool'
    NAME = 'Boolean'

    def __str__(self):
        return 'Yes' if self._value else 'No'


class Ratio(_SimpleValue):
    CLASS = 'ratio'
    NAME = 'Ratio'

    def __str__(self):
        return '{} %'.format(self._value * 100)


class Timestamp(_SimpleValue):
    CLASS = 'timestamp'
    NAME = 'Timestamp'

    def __str__(self):
        return utils.format_timestamp(self._value)


class TimeRange(_Data):
    CLASS = 'time-range'
    NAME = 'Time range'
    COLUMNS = (
        ('beginning', 'begin'),
        ('ending', 'end'),
    )

    def __init__(self, begin, end):
        super().__init__()
        self._begin = begin
        self._end = end

    @property
    def begin(self):
        return self._begin

    @property
    def end(self):
        return self._end

    @classmethod
    def from_obj(cls, obj):
        return cls(obj['begin'], obj['end'])

    def __str__(self):
        begin = utils.format_timestamp(self._begin)
        end = utils.format_timestamp(self._end)

        return '{} -> {}'.format(begin, end)


class Duration(_SimpleValue):
    CLASS = 'duration'
    NAME = 'Duration'
    UNIT = 'µs'

    def __str__(self):
        return '{:.3f}'.format(round(self._value / 1000, 3))


class Size(_SimpleValue):
    CLASS = 'size'
    NAME = 'Size'

    def __str__(self):
        return utils.format_size(self._value)


class Bitrate(_SimpleValue):
    CLASS = 'bitrate'
    NAME = 'Bitrate'
    UNIT = 'bps'


class Syscall(_SimpleName):
    CLASS = 'syscall'
    NAME = 'System call'

    def __str__(self):
        return '{}()'.format(self._name)


class Process(_Data):
    CLASS = 'process'
    NAME = 'Process'
    COLUMNS = (
        ('name', 'name'),
        ('PID', 'pid'),
        ('TID', 'tid'),
    )

    def __init__(self, name, pid, tid):
        super().__init__()
        self._name = name
        self._pid = pid
        self._tid = tid

    @property
    def name(self):
        return self._name

    @property
    def pid(self):
        return self._pid

    @property
    def tid(self):
        return self._tid

    @classmethod
    def from_obj(cls, obj):
        name = Unknown()
        pid = Unknown()
        tid = Unknown()

        if 'name' in obj:
            name = String(obj['name'])

        if 'pid' in obj:
            pid = Integer(obj['pid'])

        if 'tid' in obj:
            tid = Integer(obj['tid'])

        return cls(name, pid, tid)

    def __str__(self):
        parts = []

        if type(self._name) is not Unknown:
            parts.append(str(self._name))

        if type(self._pid) is not Unknown:
            parts.append('(PID: {})'.format(self._pid))

        if type(self._tid) is not Unknown:
            parts.append('(TID: {})'.format(self._tid))

        if not parts:
            return '?'

        return ' '.join(parts)


class Path(_Data):
    CLASS = 'path'
    NAME = 'File system path'

    def __init__(self, path):
        super().__init__()
        self._path = path

    @property
    def path(self):
        return self._path

    @classmethod
    def from_obj(cls, obj):
        return cls(obj['path'])

    def __str__(self):
        return self._path


class FileDescriptor(_Data):
    CLASS = 'fd'
    NAME = 'File descriptor'

    def __init__(self, fd):
        super().__init__()
        self._fd = fd

    @property
    def fd(self):
        return self._fd

    @classmethod
    def from_obj(cls, obj):
        return cls(obj['fd'])

    def __str__(self):
        return str(self._fd)


class Irq(_Data):
    CLASS = 'irq'
    NAME = 'Interrupt'
    COLUMNS = (
        ('hard?', 'hard'),
        ('name', 'name'),
        ('number', 'nr'),
    )

    def __init__(self, hard, nr, name):
        super().__init__()
        self._hard = hard
        self._nr = nr
        self._name = name

    @property
    def hard(self):
        return self._hard

    @property
    def nr(self):
        return self._nr

    @property
    def name(self):
        return self._name

    def __str__(self):
        parts = []

        if self._hard:
            parts.append('Hard:')
        else:
            parts.append('Soft:')

        if type(self._name) is not Unknown:
            parts.append(str(self._name))

        if type(self._nr) is not Unknown:
            parts.append('(#{})'.format(self._nr))

        return ' '.join(parts)

    @classmethod
    def from_obj(cls, obj):
        hard = True
        nr = Unknown()
        name = Unknown()

        if 'hard' in obj:
            hard = Bool(obj['hard'])

        if 'nr' in obj:
            nr = Integer(obj['nr'])

        if 'name' in obj:
            name = String(obj['name'])

        return cls(hard, nr, name)


class Cpu(_Data):
    CLASS = 'cpu'
    NAME = 'CPU'

    def __init__(self, cpu_id):
        super().__init__()
        self._cpu_id = cpu_id

    @property
    def id(self):
        return self._cpu_id

    @classmethod
    def from_obj(cls, obj):
        return cls(obj['id'])

    def __str__(self):
        return str(self._cpu_id)


class Disk(_SimpleName):
    CLASS = 'disk'
    NAME = 'Disk'


class Part(_SimpleName):
    CLASS = 'part'
    NAME = 'Disk partition'


class NetIf(_SimpleName):
    CLASS = 'netif'
    NAME = 'Network interface'


class ColumnDescription:
    def __init__(self, title, cls, unit):
        super().__init__()
        self._title = title
        self._cls = cls
        self._unit = unit

    @property
    def title(self):
        return self._title

    @property
    def cls(self):
        return self._cls

    @property
    def unit(self):
        return self._unit

    @classmethod
    def from_obj(cls, obj):
        title = None
        column_cls = None
        unit = None

        if 'title' in obj:
            title = obj['title']

        if 'class' in obj:
            class_name = obj['class']

            if class_name in _data_class_names_to_class:
                column_cls = _data_class_names_to_class[class_name]
            else:
                print(class_name)
                column_cls = class_name

        if 'unit' in obj:
            unit = obj['unit']

        return cls(title, column_cls, unit)


class TableClass:
    def __init__(self, title, column_descriptions):
        self._title = title
        self._column_descriptions = column_descriptions

    @classmethod
    def from_obj(cls, obj):
        title = None
        column_descriptions = []

        if 'title' in obj:
            title = obj['title']

        if 'column-descriptions' in obj:
            for cd_obj in obj['column-descriptions']:
                column_descriptions.append(ColumnDescription.from_obj(cd_obj))

        return cls(title, column_descriptions)

    @property
    def title(self):
        return self._title

    @property
    def column_descriptions(self):
        return self._column_descriptions


@total_ordering
class Version:
    def __init__(self, major=0, minor=0, patch=0, extra=0):
        self._major = major
        self._minor = minor
        self._patch = patch
        self._extra = extra

    @property
    def major(self):
        return self._major

    @property
    def minor(self):
        return self._minor

    @property
    def patch(self):
        return self._patch

    @property
    def extra(self):
        return self._extra

    def _to_int(self):
        return self._major * 10000 + self._minor * 100 + self._patch

    def __eq__(self, other):
        if type(other) is not type(self):
            return False

        return self._to_int() == other._to_int()

    def __lt__(self, other):
        if type(other) is not type(self):
            return False

        return self._to_int() < other._to_int()

    def __str__(self):
        base = '{}.{}.{}'.format(self._major, self._minor, self._patch)

        if self._extra is not None:
            base += '-{}'.format(self._extra)

        return base

    @classmethod
    def from_obj(cls, obj):
        major = 0
        minor = 0
        patch = 0
        extra = None

        if 'major' in obj:
            major = obj['major']

        if 'minor' in obj:
            minor = obj['minor']

        if 'patch' in obj:
            patch = obj['patch']

        if 'extra' in obj:
            extra = obj['extra']

        return cls(major, minor, patch, extra)


class Metadata:
    def __init__(self, mi_version, version, title, authors, description,
                 url, tags, table_classes=None):
        self._mi_version = mi_version
        self._version = version
        self._title = title
        self._authors = authors
        self._description = description
        self._url = url
        self._tags = tags

        if table_classes is None:
            table_classes = {}

        self._table_classes = table_classes

    @property
    def mi_version(self):
        return self._mi_version

    @property
    def version(self):
        return self._version

    @property
    def title(self):
        return self._title

    @property
    def authors(self):
        return self._authors

    @property
    def description(self):
        return self._description

    @property
    def url(self):
        return self._url

    @property
    def tags(self):
        return self._tags

    @property
    def table_classes(self):
        return self._table_classes

    @classmethod
    def from_obj(cls, obj):
        mi_version = None
        version = None
        title = None
        authors = []
        description = None
        url = None
        tags = []

        if 'mi-version' in obj:
            mi_version = Version.from_obj(obj['mi-version'])

        if 'version' in obj:
            version = Version.from_obj(obj['version'])

        if 'title' in obj:
            title = obj['title']

        if 'authors' in obj:
            authors = obj['authors']

        if 'description' in obj:
            description = obj['description']

        if 'url' in obj:
            url = obj['url']

        if 'tags' in obj:
            tags = obj['tags']

        return cls(mi_version, version, title, authors, description, url, tags)

    @property
    def dict(self):
        return self._dict


class ResultTable:
    def __init__(self, time_range, cls, rows):
        super().__init__()
        self._time_range = time_range
        self._cls = cls
        self._rows = rows

    @property
    def time_range(self):
        return self._time_range

    @property
    def cls(self):
        return self._cls

    @property
    def rows(self):
        return self._rows

    @classmethod
    def from_obj(cls, obj):
        time_range = None
        rows = []

        if 'time-range' in obj:
            time_range = TimeRange.from_obj(obj['time-range'])

        if 'data' in obj:
            for row in obj['data']:
                cells = []

                for cell_obj in row:
                    cells.append(_data_from_obj(cell_obj))

                rows.append(cells)

        return cls(time_range, None, rows)


class AnalysisResults:
    def __init__(self, result_tables, metadata):
        super().__init__()
        self._result_tables = result_tables
        self._metadata = metadata

    @property
    def result_tables(self):
        return self._result_tables

    @property
    def metadata(self):
        return self._metadata

    @property
    def dict(self):
        return self._dict


class ProgressUpdate:
    def __init__(self, value, msg):
        self._value = value
        self._msg = msg

    @property
    def value(self):
        return self._value

    @property
    def message(self):
        return self._msg

    @classmethod
    def from_line(cls, line):
        first_space_pos = line.find(' ')
        msg = None
        number = line

        if first_space_pos >= 0:
            number = line[:first_space_pos]
            msg = line[first_space_pos + 1:]

        try:
            number = float(number)
        except:
            return

        return cls(number, msg)


_data_classes = (
    Integer,
    Number,
    String,
    Bool,
    Unknown,
    Ratio,
    Timestamp,
    TimeRange,
    Duration,
    Size,
    Bitrate,
    Syscall,
    Process,
    Path,
    FileDescriptor,
    Irq,
    Cpu,
    Disk,
    Part,
    NetIf,
)


_data_class_names_to_class = {cls.CLASS: cls for cls in _data_classes}


def _data_from_obj(obj):
    if type(obj) is dict and 'class' in obj:
        cls = _data_class_names_to_class[obj['class']]
    else:
        if type(obj) is int:
            return Integer(obj)
        elif type(obj) is float:
            return Number(obj)
        elif type(obj) is str:
            return String(obj)
        elif type(obj) is bool:
            return Bool(obj)
        elif type(obj) is None:
            return Empty()
        else:
            return Unknown()

    return cls.from_obj(obj)


def _table_class_from_obj(obj, table_classes):
    table_class = TableClass.from_obj(obj)

    if 'inherit' in obj:
        parent = table_classes[obj['inherit']]

        if table_class.title is None:
            table_class._title = parent.title

        if not table_class.column_descriptions:
            table_class._column_descriptions = parent.column_descriptions

    return table_class


def _metadata_from_obj(obj):
    metadata = Metadata.from_obj(obj)
    table_classes = {}

    if 'table-classes' not in obj:
        return metadata

    for table_class_name, table_class_obj in obj['table-classes'].items():
        table_classes[table_class_name] = _table_class_from_obj(table_class_obj,
                                                                table_classes)

    metadata._table_classes = table_classes

    return metadata


def _result_table_from_obj(obj, table_classes):
    result_table = ResultTable.from_obj(obj)

    if type(obj['class']) is str:
        table_class = table_classes[obj['class']]
    else:
        table_class = _table_class_from_obj(obj['class'], table_classes)

    result_table._cls = table_class

    return result_table


def _analysis_results_from_obj(obj, metadata):
    result_tables = []

    for result_obj in obj['results']:
        result_table = _result_table_from_obj(result_obj,
                                              metadata.table_classes)
        result_tables.append(result_table)

    return AnalysisResults(result_tables, metadata)


def _error_or_other_from_obj(obj, other_fn):
    if 'error-code' in obj or 'error-message' in obj:
        model_obj = Error.from_obj(obj)
    else:
        model_obj = other_fn(obj)

    model_obj._dict = obj

    return model_obj


def get_metadata(cmd):
    full_cmd = '{} --metadata'.format(cmd)
    _logger.info('Getting metadata with full command: {}'.format(full_cmd))

    try:
        output = subprocess.check_output(full_cmd, shell=True,
                                         stderr=subprocess.STDOUT)
    except subprocess.CalledProcessError as e:
        _logger.error('Cannot get metadata: command returned {}'.format(e.returncode))
        raise ExecError(e.returncode, e.output) from e
    except Exception as e:
        _logger.error('Cannot get metadata')
        raise ExecError() from e

    _logger.info('Parsing metadata')

    try:
        output = output.decode()
        json_data = json.loads(output)
    except Exception as e:
        _logger.error('Cannot parse metadata')
        raise ParseError('Cannot parse metadata') from e

    if 'mi-version' in json_data:
        mi_version = Version.from_obj(json_data['mi-version'])

        if mi_version > Version(0, 1):
            raise UnsupportedVersion('Unsupported MI version: {}'.format(mi_version))

    try:
        ret = _error_or_other_from_obj(json_data, _metadata_from_obj)
    except Exception as e:
        _logger.error('Cannot parse metadata')
        raise ParseError('Cannot parse metadata') from e

    return ret


def get_analysis_results(cmd, path, begin, end, limit, output_progress,
                         on_progress_update=None):
    def analysis_results_from_obj(obj):
        return _analysis_results_from_obj(obj, metadata)

    full_cmd = cmd

    if begin is not None:
        full_cmd += ' --begin={}'.format(begin)

    if end is not None:
        full_cmd += ' --end={}'.format(end)

    if limit is not None:
        full_cmd += ' --limit={}'.format(limit)

    if output_progress:
        full_cmd += ' --output-progress'

    full_cmd += " '{}'".format(path)
    metadata = get_metadata(cmd)

    if type(metadata) is Error:
        return metadata

    json_lines = []
    in_progress = True
    _logger.info('Getting analysis results with full command: {}'.format(full_cmd))

    try:
        p = subprocess.Popen(full_cmd, shell=True, stderr=subprocess.PIPE,
                             stdout=subprocess.PIPE, bufsize=1,
                             universal_newlines=True)

        with p.stdout:
            for line in p.stdout:
                if line.endswith('\n'):
                    line = line[:-1]

                if in_progress:
                    if len(line.strip()) == 0:
                        continue

                    progress_update = ProgressUpdate.from_line(line)

                    if progress_update is None:
                        in_progress = False
                    else:
                        if on_progress_update is not None:
                            on_progress_update(progress_update)

                if not in_progress:
                    json_lines.append(line)

        _logger.info('Waiting for the command to complete')
        p.wait()
    except Exception as e:
        import traceback
        traceback.print_exc()
        _logger.error('Cannot get analysis results')
        raise ExecError() from e

    _logger.info('Parsing analysis results')

    try:
        output = '\n'.join(json_lines)
        json_data = json.loads(output)
        ret = _error_or_other_from_obj(json_data, analysis_results_from_obj)
    except Exception as e:
        _logger.error('Cannot parse analysis results')
        raise ParseError('Cannot parse analysis results') from e

    return ret



# The MIT License (MIT)
#
# Copyright (c) 2016 Philippe Proulx <pproulx@efficios.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from pkg_resources import resource_filename
from PyQt5 import Qt, QtGui, uic
import os.path


class QCommonDialog(Qt.QDialog):
    def __init__(self):
        super().__init__()

    def show_move(self, pos):
        self.move(pos)
        self.exec()


class QtUiLoad:
    def _load_ui(self):
        ui_rel_path = os.path.join('res', 'ui', '{}.ui'.format(self._UI_NAME))
        ui_path = resource_filename(__name__, ui_rel_path)
        uic.loadUi(ui_path, baseinstance=self)


def get_res_path(name):
    rel_path = os.path.join('res', name)
    path = resource_filename(__name__, rel_path)

    return path


def get_qicon(name):
    filename = '{}.png'.format(name)
    path = get_res_path(os.path.join('icons', filename))

    return Qt.QIcon(path)


def set_widget_icon(widget, name, size=16):
    icon = get_qicon(name)
    widget.setIcon(icon)
    widget.setIconSize(Qt.QSize(size, size))


def set_widget_icons(entries, size=16):
    for entry in entries:
        set_widget_icon(entry[0], entry[1], size)


def goto_url(url):
    QtGui.QDesktopServices.openUrl(Qt.QUrl(url))


def usec_to_sec(usec):
    return int(usec) / 1000000


def sec_to_usec(sec):
    return round(sec * 1000000)


def bytes_to_human_prefix(b):
    if b < (1 << 10):
        return (b, 'b')
    elif b < (1 << 20):
        return (b / (1 << 10), 'k')
    elif b < (1 << 30):
        return (b / (1 << 20), 'm')
    else:
        return (b / (1 << 30), 'g')


def bytes_to_human(b):
    prefix_map = {
        'b': 'B',
        'k': 'kiB',
        'm': 'MiB',
        'g': 'GiB',
    }
    b, prefix = bytes_to_human_prefix(b)

    return '{} {}'.format(b, prefix_map[prefix])


# Copyright (C) 2016 - Antoine Busque <abusque@efficios.com>
# MIT license
def format_timestamp(timestamp, print_date=False, gmt=False):
    import time

    NSEC_PER_SEC = 1000000000

    """Format a timestamp into a human-readable date string

    Args:
        timestamp (int): nanoseconds since epoch.

        print_date (bool, optional): flag indicating whether to print
        the full date or just the time of day (default: False).

        gmt (bool, optional): flag indicating whether the timestamp is
        in the local timezone or gmt (default: False).

    Returns:
        The formatted date string, containing either the full date or
        just the time of day.
    """
    date_fmt = '{:04}-{:02}-{:02} '
    time_fmt = '{:02}:{:02}:{:02}.{:09}'

    if gmt:
        date = time.gmtime(timestamp / NSEC_PER_SEC)
    else:
        date = time.localtime(timestamp / NSEC_PER_SEC)

    formatted_ts = time_fmt.format(
        date.tm_hour, date.tm_min, date.tm_sec,
        timestamp % NSEC_PER_SEC
    )

    if print_date:
        date_str = date_fmt.format(date.tm_year, date.tm_mon, date.tm_mday)
        formatted_ts = date_str + formatted_ts

    return formatted_ts


# Copyright (C) 2016 - Antoine Busque <abusque@efficios.com>
# MIT license
def format_size(size, binary_prefix=True):
    import math

    """Convert an integral number of bytes to a human-readable string.

    Args:
        size (int): a non-negative number of bytes.

        binary_prefix (bool, optional): whether to use binary units
        prefixes, over SI prefixes (default: True).

    Returns:
        The formatted string comprised of the size and units.

    Raises:
        ValueError: if size < 0.
    """
    if size < 0:
        raise ValueError('Cannot format negative size')

    if binary_prefix:
        base = 1024
        units = ['  B', 'kiB', 'MiB', 'GiB', 'TiB', 'PiB', 'EiB', 'ZiB', 'YiB']
    else:
        base = 1000
        units = [' B', 'KB', 'MB', 'GB', 'TB', 'PB', 'EB', 'ZB', 'YB']

    if size == 0:
        exponent = 0
    else:
        exponent = int(math.log(size, base))
        if exponent >= len(units):
            # Don't try and use a unit above YiB/YB
            exponent = len(units) - 1

        size /= math.pow(base, exponent)

    unit = units[exponent]

    if exponent == 0:
        # Don't display fractions of a byte
        format_str = '{:0.0f} {}'
    else:
        format_str = '{:0.2f} {}'

    return format_str.format(size, unit)
